<copyright> MU Compound.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA

    <id>$Id: Compound.t,v 1.16 1999/09/22 19:53:22 tiggr Exp $</id>
    </copyright>

/******************** Context ********************/

// Deprecated?
// Sun Sep  5 11:11:20 1999, tiggr@gerbil.org
<doc> A Context is something that can be searched for variables and the
    like.  </doc>
implementation class
Context: Common
end;

implementation instance
Context

<doc> Return the enclosing block, or {nil}.  </doc>
deferred BlockDefinition
  block;

<doc> Return the enclosing context.  </doc>
deferred Context
  container;

<doc> Search the receiving context for a local variable or argument.
    Default implementation passes this to the {container}.  </doc>
VariableSingle (v)
  localEntityNamed String name
{
  = [[self container] localEntityNamed name];
}

<doc> Return the method containing the receiving compound.  Default
    implementation passes this to the {container}.  </doc>
CustomMethod
  method
{
  = [[self container] method];
}

end;

/******************** Compound ********************/

<doc> A Compound is something that holds a bunch of expressions and
    declarations.  It is itself also an expression.  </doc>
implementation class
Compound: Context, Expression
end;

implementation instance
Compound
{
  <doc> The containing compound or method, container-wise.  </doc>
  public Context container;

  <doc> The local variables, keyed on their name.  </doc>
  MutableDictionary variables;

  <doc> The local variables, in order.  This is also faster when visiting
      all variables.  </doc>
  MutableArray variables_ordered;

  <doc> The expressions in this compound.  </doc>
  MutableArray expressions;

  <doc> The last expression.  Not necessarily equals the last entry in
      {expressions}.  It could be the expression value resulting from a
      local variable declaration (which is why it is mutable).  </doc>
  mutable Expression last_expression;
}

<doc> Designated initializer.  </doc>
id
  initWithContainer Context c
{
  container = c;
  expressions = [MutableObjectArray new];
  = [super initWithType nil];
}

<doc> Add the {top_expression} to the {expressions}, and update the
    {last_expression}.  </doc>
void
  addExpression Expression top_expression
{
  [expressions add top_expression];
  last_expression = top_expression;
}

<doc> Note the new local {variable}.  </doc>
void
  addVariable LocalVariable variable
pre
  [variable container] == self
{
  String vname = [variable name];

  if (!variables)
    {
      variables = [MutableDictionary new];
      variables_ordered = [MutableObjectArray new];
    }
  else
    {
      LocalVariable v = variables[vname];
      if (v != nil)
	{
	  [[[Parser error variable] print ("redeclaration of `", vname, "'")]
	    nl];
	  [[[Parser report v] print ("`", vname, "' previously declared here")]
	    nl];
	}
    }
  variables[vname] = variable;
  [variables_ordered add variable];
}

<doc> Ask the {container}.  </doc>
BlockDefinition
  block
{
  = [container block];
}

<doc> Search for an entity with the given name.  If not found, pass to the
    container.  </doc>
VariableSingle (lv)
  localEntityNamed String name
{
  if (variables != nil)
    lv = variables[name];
  if (!lv)
    lv = [container localEntityNamed name];
}

<doc> Create a new local variable in the current compound.  </doc>
LocalVariable (lv)
  localVariable (Type, String) (t, n)
	    for Common decl
{
  lv = !decl ? [LocalVariable alloc] : [LocalVariable alloc decl];
  lv = [lv initWithName n type t container self];
  [self addVariable lv];
}

<doc> Ask the {container}.  </doc>
CustomMethod
  method
{
  = [container method];
}

<doc> Return the {variables_ordered}.  </doc>
Indexed
  variables
{
  = variables_ordered;
}

end;

/******************** MethodBody ********************/

<doc> A MethodBody is the compound used as the body of a method.  </doc>
implementation class
MethodBody: Compound

end;

implementation instance
MethodBody
{
  <doc> Our container (sort-of) is a method.  </doc>
  redeclare CustomMethod container;
}

<doc> Apparently, there is no block between the compound propagating this
    question and this method body.  Return {nil}.  </doc>
BlockDefinition
  block
{
  = nil;
}

<doc> Redeclare our {container} accessor.  </doc>
redeclare CustomMethod
  container;

<doc> Return the method containing the receiving compound, which is the
    {container}.  </doc>
CustomMethod
  method
{
  = container;
}

end;

/******************** BlockDefinition ********************/

implementation class
BlockDefinition: Compound, Callable

end;

implementation instance
BlockDefinition
{
  <doc> The block variables, that are local to this block and remain their
      value through invocations.  </doc>
  MutableDictionary block_variables;
}

<doc> Return {self}.  </doc>
BlockDefinition
  block
{
  = self;
}

<doc> Note the block variable with the indicated {type}, {name}, and
    {value}.  The {value} may be {nil} to indicate absence of an
    initial-value assignment.  </doc>
void
 blockVariable (Type, String, Expression) (type, name, value)
{
  if (!block_variables)
    block_variables = [MutableDictionary new];
  block_variables[name] = [[BlockVariable alloc] initWithName name type type
						 value value block self];
}

<doc> The container of a block is a compound.  </doc>
redeclare Compound
  container;

<doc> Match with the arguments first.  </doc>
VariableSingle (v)
  localEntityNamed String name
{
  v = [self argumentNamed name];
  if (!v)
    {
      if (block_variables != nil)
	v = block_variables[name];

      if (!v)
	{
	  v = [[self container] localEntityNamed name];
	  if (v != nil)
	    v = [self variableProxy v];
	}
    }
}

<doc> Return the proxy for {v} from the enclosing context.  </doc>
VariableSingle (v)
  variableProxy VariableSingle v
{
}

end;

/******************** BlockBody ********************/

<doc> A {BlockBody} is to a {BlockDefinition} what the {MethodBody} is to
    a {CustomMethod}.  </doc>
implementation class
BlockBody: Compound

end;

implementation instance
BlockBody
{
  <doc> Our {container} is a {BlockDefinition}.  </doc>
  redeclare BlockDefinition container;
}

<doc> Return our {container} block.  </doc>
BlockDefinition
  block
{
  = container;
}

<doc> The container of a block body is a block definition.  </doc>
redeclare BlockDefinition
  container;

end;
