<copyright> MU extension.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1996-1998 Pieter J. Schoenmakers.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA

    <id>$Id: Extension.t,v 1.8 1998/07/27 15:20:56 tiggr Exp $</id>
    </copyright>

implementation class
Extension: Named

<doc> Return a newly allocated {Extension} which extends the {meta}.  The
    extension is contained in the {file}, and has the {name}.  Upon
    return, the extension has been added to the meta.  </doc>
instance (id) (ext)
  ofMeta Meta meta
  inFile TomFile file
   name: String name = nil
{
  ext = [[self alloc] initWithMeta meta inFile file name name];
  [meta addExtension ext];
}

end;

implementation instance
Extension
{
  <doc> The meta we extend.  </doc>
  public Meta meta;

  <doc> The file in which containing this extension.  </doc>
  public TomFile file;

  <doc> Our descriptive name, created on demand.  </doc>
  String descriptive_name;

  <doc> The non-static variables in this extension.  </doc>
  MutableArray variables;

  <doc> Method definitions in this extension.  </doc>
  MutableArray methods;

  <doc> Any attributes in this extension that are not in the {variables}.
      </doc>
  MutableArray other_attributes;
}

<doc> Designated initializer.  </doc>
private id
  initWithMeta Meta m
	inFile TomFile f
	  name String n
{
  (meta, file) = (m, f);
  = [super initWithName n];
}

<doc> Add the alias {a} to the {other_attributes} and inform the {meta}.
    </doc>
void
  addAlias Alias a
{
  if (!other_attributes)
    other_attributes = [MutableObjectArray with a];
  else
    [other_attributes add a];

  [meta alias a in self];
}

<doc> Be notified of the encounter of a const.  </doc>
void
  addConst (String, Expression) (name, value)
{
  Constant c = [[Constant alloc] initWithName name value value extension self];

  if (!other_attributes)
    other_attributes = [MutableObjectArray with c];
  else
    [other_attributes add c];

  [meta constant c in self];
}

<doc> Be notified of the encounter of a class or instance variable, or a
    redeclaration thereof.  </doc>
void
  addMetaVariable (int, Type, String) (qualifiers, type, name)
{
  Attribute a;

  if (![Qualified isRedeclare qualifiers])
    a = [[MetaVariable alloc] initWithName name qualifiers qualifiers
			      type type extension self];
  else
    {
      a = [[Alias alloc] initWithName name type type extension self];
      if ([Qualified set qualifiers isRedeclare NO] != 0)
	[[[Parser error a] print "redeclare voids other qualifiers"] nl];
    }

  [a register];
}

<doc> Add the {method} to the {methods}.  </doc>
void
  addMethod Method method
{
  if (!methods)
    methods = [MutableObjectArray new];
  [methods add method];
}

<doc> Add the variable {v} to the {variables} or, if it is static, to the
    {other_attributes}.  </doc>
void
  addVariable MetaVariable v
{
  if ([v isStatic])
    if (!other_attributes)
      other_attributes = [MutableObjectArray with v];
    else
      [other_attributes add v];
  else if (!variables)
    variables = [MutableObjectArray with v];
  else
    [variables add v];

  [meta variable v in self];
}

<doc> Return the {descriptive_name}, creating it if necessary.  </doc>
String
  descriptive_name
{
  if (!descriptive_name)
    {
      MutableByteString s = [[MutableByteString new]
			      print ([meta isClass] ? '+' : '-',
				     [[[[meta extensionNamed nil]
					 file] unit] name], '.', [meta name])];
      if (name != nil)
	[s print ('(', name, ')')];
      [s freeze];
      descriptive_name = s;
    }

  = descriptive_name;
}

<doc> Return {TRUE} if our {name} is {nil}.  </doc>
boolean
  isMainExtension
{
  = !name;
}

<doc> Return the unit containing this extension.  </doc>
Unit
  unit
{
  = [file unit];
}

end;
