<copyright> MU lexer.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright (C) 1996 Pieter J. Schoenmakers.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA

    <id>$Id: Lexer.t,v 1.13 1998/07/17 23:33:07 tiggr Exp $</id>
    </copyright>

<doc> The {Lexer} class is a generic lexer for tom sources, not tied to a
    particular parser.  Subclasses of it are used by the tom and unit
    parsers.  </doc>

implementation class
Lexer: gps.StreamLexer
{
  <doc> The set of identifier-like tokens known by this lexer class.  </doc>
  MutableDictionary tokens;
}

<doc> A redefinition of {alloc} to provide some of Objective-C's
    +initialize functionality.  </doc>
redefine instance (id)
  alloc
{
  if (!tokens)
    {
      tokens = [MutableDictionary new];
      [self fillTokens];
    }

  = [super (class (StreamLexer)) alloc];
}

<doc> Add the identifier {name} to the identifier-like {tokens} as the
    {Token} value {number}.  </doc>
void
  addToken (ByteString, int) (name, number)
{
  UniqueString u = [UniqueString with name];

  tokens[u] = [Token with u token number];
}

<doc>  Fill the {tokens} by repeatedly invoking {addToken}.  </doc>
deferred void
  fillTokens;

deferred class (Parser)
  parserClass;

end;

implementation instance
Lexer
{
  <doc> The buffer for what we read.  </doc>
  MutableByteString buf;

  <doc> The (cached) size of the buffer.  </doc>
  int size;

  <doc> The start position of the current token in the buffer, the next
      position in the buffer, and the end position (one beyond last).  </doc>
  int start, next, beyond;

  <doc> The semantics of an integer kind.  </doc>
  int intvalue;

  <doc> A substring used to refer to the documentation tag in the {buf}.
      </doc>
  ByteSubstring doc_tag;

  <doc> Iff {TRUE}, we are supposedly interested in documentation.  </doc>
  public mutable boolean docp;

  <doc> Iff {TRUE}, we do not collect semantics with the tokens.  </doc>
  public mutable boolean skips;
}

redefine id
  initWithStream InputStream s
	   named String n
{
  buf = [MutableByteString withCapacity size = 8192];
  [buf resize size];
  doc_tag = [ByteSubstring alloc];

  = [super initWithStream s named n];
}

<doc> Return {TRUE} iff we are interested in documentation that starts
    with the {tag}.  The default implementation simply asks self whether
    {docp} (plain).  </doc>
boolean
  docp String tag
{
  = [self docp];
}

extern int
  nextToken;

int
  semantics
{
  = intvalue;
}

<doc> Various methods to return the correct token number for a specific
    parser.  </doc>
deferred int character;
deferred int identifier;
deferred int number;
deferred int string;

<doc> The default action for returning documentation is to skip it.  This
    is necessary if {docp} is set while lexer thinks it should not.  A
    lexer genuinly interested in documentation, for the time being, should
    return the correct documentation token to the caller, or it could
    inform the {parser} of the documentation, which avoids having to flood
    the grammar with DOCUMENTATION tokens.  </doc>
int
  documentation String body
	    tag String tag
	     at int line
{
  = [self nextToken];
}

// Sat Jul 18 00:22:52 1998, tiggr@gerbil.org
// We need a `obsolete' keyword...
int
  documentation String body
	    tag String tag
{
  = [self documentation body tag tag at current_line];
}

end;
