<copyright> MU methods
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1997 Pieter J. Schoenmakers.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA

    <id>$Id: Method.t,v 1.22 1999/09/25 21:29:29 tiggr Exp $</id>
    </copyright>

/******************** Method ********************/

implementation class
Method: Common, mu

end;

implementation instance
Method
{
  <doc> The type returned by this method.  </doc>
  public Type return_type;
}

<doc> Designated initializer.  </doc>
id
  initWithType Type rt
{
  return_type = rt;
  = [super init];
}

<doc> Return the number of name parts.  </doc>
deferred int
  namePartsLength;

end;

/******************** Builtinmethod ********************/

implementation class
BuiltinMethod: Method

<doc> Add all builtin methods (i.e., the operators) to the {top_type}.
    </doc>
void
  initializeOperators
{
  int i;

  Type bt = basic_types_from_kind[BT_BYTE];
  Type ot = basic_types_from_kind[BT_BOOLEAN];
  Type ct = basic_types_from_kind[BT_CHAR];
  Type it = basic_types_from_kind[BT_INT];
  Type lt = basic_types_from_kind[BT_LONG];
  Type ft = basic_types_from_kind[BT_FLOAT];
  Type dt = basic_types_from_kind[BT_DOUBLE];

  // Can't we create this, dump the objects to a file, and reload them?
  // Thu Apr 30 01:29:56 1998, tiggr@gerbil.org
  [self monop OP_MIN argType it returnType it];
  [self monop OP_MIN argType lt returnType lt];
  [self monop OP_MIN argType ft returnType ft];
  [self monop OP_MIN argType dt returnType dt];
			    
  [self monop OP_INV argType ot returnType ot];
  [self monop OP_INV argType it returnType it];
  [self monop OP_INV argType lt returnType lt];

  [self monop OP_PP argType bt returnType bt];
  [self monop OP_PP argType ct returnType ct];
  [self monop OP_PP argType it returnType it];
  [self monop OP_PP argType lt returnType lt];
  [self monop OP_PP argType ft returnType ft];
  [self monop OP_PP argType dt returnType dt];

  [self monop OP_MM argType bt returnType bt];
  [self monop OP_MM argType ct returnType ct];
  [self monop OP_MM argType it returnType it];
  [self monop OP_MM argType lt returnType lt];
  [self monop OP_MM argType ft returnType ft];
  [self monop OP_MM argType dt returnType dt];

  for (i = BT_BOOLEAN; i < BT_DYNAMIC; i++)
    [self monop OP_NOT argType basic_types_from_kind[i] returnType ot];
  [self monop OP_NOT argType top_type returnType ot];

  for (i = BT_BOOLEAN; i < BT_DYNAMIC; i++)
    {
      Type t = basic_types_from_kind[i];
      [self monop OP_OLD argType t returnType t];
    }
  [self monop OP_OLD argType top_type returnType top_type];

  [self binop OP_MUL argTypes (bt, bt) returnType bt];
  [self binop OP_MUL argTypes (ct, ct) returnType ct];
  [self binop OP_MUL argTypes (it, it) returnType it];
  [self binop OP_MUL argTypes (lt, lt) returnType lt];
  [self binop OP_MUL argTypes (ft, ft) returnType ft];
  [self binop OP_MUL argTypes (dt, dt) returnType dt];

  [self binop OP_DIV argTypes (bt, bt) returnType bt];
  [self binop OP_DIV argTypes (ct, ct) returnType ct];
  [self binop OP_DIV argTypes (it, it) returnType it];
  [self binop OP_DIV argTypes (lt, lt) returnType lt];
  [self binop OP_DIV argTypes (ft, ft) returnType ft];
  [self binop OP_DIV argTypes (dt, dt) returnType dt];

  [self binop OP_MOD argTypes (bt, bt) returnType bt];
  [self binop OP_MOD argTypes (ct, ct) returnType ct];
  [self binop OP_MOD argTypes (it, it) returnType it];
  [self binop OP_MOD argTypes (lt, lt) returnType lt];

  [self binop OP_ADD argTypes (bt, bt) returnType bt];
  [self binop OP_ADD argTypes (ct, ct) returnType ct];
  [self binop OP_ADD argTypes (it, it) returnType it];
  [self binop OP_ADD argTypes (lt, lt) returnType lt];
  [self binop OP_ADD argTypes (ft, ft) returnType ft];
  [self binop OP_ADD argTypes (dt, dt) returnType dt];

  [self binop OP_SUB argTypes (bt, bt) returnType bt];
  [self binop OP_SUB argTypes (ct, ct) returnType ct];
  [self binop OP_SUB argTypes (it, it) returnType it];
  [self binop OP_SUB argTypes (lt, lt) returnType lt];
  [self binop OP_SUB argTypes (ft, ft) returnType ft];
  [self binop OP_SUB argTypes (dt, dt) returnType dt];

  [self binop OP_SHL argTypes (it, it) returnType it];
  [self binop OP_SHL argTypes (lt, it) returnType lt];

  [self binop OP_SHR argTypes (it, it) returnType it];
  [self binop OP_SHR argTypes (lt, it) returnType lt];

  [self binop OP_LSR argTypes (it, it) returnType it];
  [self binop OP_LSR argTypes (lt, it) returnType lt];

  [self binop OP_AND argTypes (ot, ot) returnType ot];
  [self binop OP_AND argTypes (bt, bt) returnType bt];
  [self binop OP_AND argTypes (ct, ct) returnType ct];
  [self binop OP_AND argTypes (it, it) returnType it];
  [self binop OP_AND argTypes (lt, lt) returnType lt];

  [self binop OP_OR argTypes (ot, ot) returnType ot];
  [self binop OP_OR argTypes (bt, bt) returnType bt];
  [self binop OP_OR argTypes (ct, ct) returnType ct];
  [self binop OP_OR argTypes (it, it) returnType it];
  [self binop OP_OR argTypes (lt, lt) returnType lt];

  [self binop OP_EOR argTypes (ot, ot) returnType ot];
  [self binop OP_EOR argTypes (bt, bt) returnType bt];
  [self binop OP_EOR argTypes (ct, ct) returnType ct];
  [self binop OP_EOR argTypes (it, it) returnType it];
  [self binop OP_EOR argTypes (lt, lt) returnType lt];

  [self binop OP_LT argTypes (bt, bt) returnType ot];
  [self binop OP_LT argTypes (ct, ct) returnType ot];
  [self binop OP_LT argTypes (it, it) returnType ot];
  [self binop OP_LT argTypes (lt, lt) returnType ot];
  [self binop OP_LT argTypes (ft, ft) returnType ot];
  [self binop OP_LT argTypes (dt, dt) returnType ot];

  [self binop OP_LE argTypes (bt, bt) returnType ot];
  [self binop OP_LE argTypes (ct, ct) returnType ot];
  [self binop OP_LE argTypes (it, it) returnType ot];
  [self binop OP_LE argTypes (lt, lt) returnType ot];
  [self binop OP_LE argTypes (ft, ft) returnType ot];
  [self binop OP_LE argTypes (dt, dt) returnType ot];

  for (i = BT_BOOLEAN; i < BT_DYNAMIC; i++)
    {
      Type bt = basic_types_from_kind[i];
      [self binop OP_NE argTypes (bt, bt) returnType ot];
    }
  [self binop OP_NE argTypes (top_type, top_type) returnType ot];

  for (i = BT_BOOLEAN; i < BT_DYNAMIC; i++)
    {
      Type bt = basic_types_from_kind[i];
      [self binop OP_EQ argTypes (bt, bt) returnType ot];
    }
  [self binop OP_EQ argTypes (top_type, top_type) returnType ot];

  [self binop OP_GE argTypes (bt, bt) returnType ot];
  [self binop OP_GE argTypes (ct, ct) returnType ot];
  [self binop OP_GE argTypes (it, it) returnType ot];
  [self binop OP_GE argTypes (lt, lt) returnType ot];
  [self binop OP_GE argTypes (ft, ft) returnType ot];
  [self binop OP_GE argTypes (dt, dt) returnType ot];

  [self binop OP_GT argTypes (bt, bt) returnType ot];
  [self binop OP_GT argTypes (ct, ct) returnType ot];
  [self binop OP_GT argTypes (it, it) returnType ot];
  [self binop OP_GT argTypes (lt, lt) returnType ot];
  [self binop OP_GT argTypes (ft, ft) returnType ot];
  [self binop OP_GT argTypes (dt, dt) returnType ot];

  [self binop OP_SC_AND argTypes (ot, ot) returnType ot];
  [self binop OP_SC_OR argTypes (ot, ot) returnType ot];
  [self binop OP_IMPLIES argTypes (ot, ot) returnType ot];
}

<doc> Hard worker for binary operators to {initializeOperators}.  </doc>
protected void
       binop int op_kind
    argTypes (Type, Type) (at1, at2)
  returnType Type rt
{
  BuiltinMethod m = [[BuiltinBinop alloc] initWithOperator op_kind returnType rt
					  argTypes (at1, at2)];
  [[top_type extensionNamed nil] addMethod m];
}

<doc> Hard worker for unary operators to {initializeOperators}.  </doc>
protected void
       monop int op_kind
     argType Type at1
  returnType Type rt
{
  BuiltinMethod m = [[BuiltinMethod alloc]
		      initWithOperator op_kind returnType rt argType at1];
  [[top_type extensionNamed nil] addMethod m];
}

end;

implementation instance
BuiltinMethod
{
  <doc> The {BT_*} thing.  </doc>
  public int operator;

  <doc> The type of the first or only argument.  </doc>
  public Type arg1;
}

<doc> Designated initializer.  </doc>
id
  initWithOperator int op
	returnType Type rt
	   argType Type at
{
  (operator, arg1) = (op, at);
  = [super initWithType rt];
}	

<doc> Return the type of the second argument, or {nil} if this is a monop.
    </doc>
Type
  arg2
{
}

<doc> Return the {top_type}.  </doc>
Meta
  meta
{
  = top_type;
}

<doc> Return 1.  </doc>
int
  namePartsLength
{
  = 1;
}

<doc> Also write {operator} and {arg1}.  </doc>
OutputStream
  writeFields OutputStream s
{
  = [[super writeFields s] print (" op=", String (operator_names[operator]),
				  " arg1=", arg1)];
}

end;

/******************** BuiltinBinop ********************/

implementation class
BuiltinBinop: BuiltinMethod

end;

implementation instance
BuiltinBinop
{
  <doc> The type of the second argument.  </doc>
  public Type arg2;
}

<doc> Designated initializer.  </doc>
id
  initWithOperator int op
	returnType Type rt
	  argTypes (Type, Type) (at1, at2)
{
  arg2 = at2;
  = [super initWithOperator op returnType rt argType at1];
}	

<doc> Return 2.  </doc>
int
  namePartsLength
{
  = 2;
}

<doc> Also write {arg2}.  </doc>
OutputStream
  writeFields OutputStream s
{
  = [[super writeFields s] print (" arg2=", arg2)];
}

end;

/******************** CustomMethod ********************/

implementation class
CustomMethod: Qualified, Context, Method, Callable
end;

implementation instance
CustomMethod
{
  <doc> The extension containing this method.  </doc>
  Extension ext;

  <doc> The name-parts of the name of this method.  </doc>
  MutableArray name_parts;

  <doc> The names assigned to the return values.  This will be a tuple of
      (tuples and) {ArgumentSingle} and {Variable} objects, or {nil} if
      not set.  </doc>
  public ArgumentTuple return_names;

  <doc> The postcondition for this method, as introduced by this method.
      </doc>
  public mutable Expression postcondition;

  <doc> Similarly, the precondition.  </doc>
  public mutable Expression precondition;
}

<doc> Designated initializer.  </doc>
id
	init Type t
  qualifiers int q
   extension Extension e
{
  name_parts = [MutableObjectArray new];
  ext = e;

  argument_singles = [MutableDictionary new];
  ArgumentSingle a = [[ArgumentSingle alloc]
		       initWithName [UniqueString with "self"]
		       type basic_types["id"]
		       callable self];
  argument_singles[[a name]] = a;
  a = [[ArgumentSingle alloc] initWithName [UniqueString with "cmd"]
			      type basic_types["selector"]
			      callable self];
  argument_singles[[a name]] = a;

  [super initWithType t];
  = [super initWithQualifiers q];
}

<doc> Return the extension {ext}.  </doc>
Extension
  extension
{
  = ext;
}

<doc> Return the {meta} of the extension {ext}.  </doc>
Meta
  meta
{
  = [ext meta];
}

<doc> Return the number of name parts.  </doc>
int
  namePartsLength
{
  = [name_parts length];
}


/********** creation **********/

void
  addNamePart String n
    argument: Argument a = nil
{
  [name_parts add n];

  if (a != nil)
    [self addArgument a];
}

<doc> Add a precondition.  We only support one of them for the moment.  </doc>
void
  addPrecondition Expression e
pre
  !precondition
{
  precondition = e;
}

<doc> Add a postcondition.  We only support one of them for the moment.  </doc>
void
  addPostcondition Expression e
pre
  !postcondition
{
  postcondition = e;
}

void
  setReturnValueNames ArgumentTuple t
{
  [t resolveReturnNamesInMethod self];

  return_names = t;
}


ArgumentSingle (arg)
  argumentNamed String name
{
  arg = [super argumentNamed name];
  if (!arg && return_names != nil) // && !body
    arg = [return_names returnValueNamed name];
}

<doc> Return the argument with the {name}, or {nil} otherwise.  </doc>
VariableSingle
  localEntityNamed String name
{
  = [self argumentNamed name];
}

/********** miscellaneous **********/

Indexed
  name_parts
{
  = name_parts;
}

<doc> Return a reference to the {return_names}.  Do not return a singleton
    tuple.  The default implementation simply returns {nil}.  </doc>
Expression
  return_value
{
}

end;
