/* C routines for mu.
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA

   $Id: glue.c,v 1.34 1999/09/27 20:19:52 tiggr Exp $  */

#include "mu-r.h"
#include <tom/util.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

/* XXX Yuck!  The buffer from which we're currently reading.  */
tom_byte *mu_buf;

/* !0 for every character allowable within an identifier.  It should be
   indexed with offset +1 to not have to check explicitly for EOF (-1).  */
char mu_identch[] =
{
  0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0,
  0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1,
  0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

tom_int
mu_premature_eof (struct _es_i_mu_Lexer *mu,
		  struct _es_i_gps_SkeletonLexer *gps,
		  int start_line, char *what, int what_len)
{
  /* XXX */
  abort ();
}

int
mu_fillbuf (struct _es_i_mu_Lexer *mu, struct _es_i_gps_StreamLexer *gps)
{
  tom_int i;

  if (mu->start)
    {
      memmove (mu_buf, mu_buf + mu->start, mu->beyond - mu->start);
      mu->beyond -= mu->start;
      mu->next -= mu->start;
      mu->start = 0;
    }
  else if (mu->beyond == mu->size)
    {
      mu->size += 8192;
      TRT_SEND (, mu->buf, SEL (v_resize_i), mu->size);
      mu_buf = TRT_SEND ((pointer_imp), mu->buf,
			 USEL (tom, _pi__byteStringContents), &i);
    }

  i = TRT_SEND (, gps->stream, USEL (tom, i_readRange__ii__into_r),
		mu->beyond, mu->size - mu->beyond, mu->buf);
  if (!i)
    return -1;

  mu->beyond += i;
  return mu_buf[mu->next++];
}

static __inline__ int
nextch (struct _es_i_mu_Lexer *mu, struct _es_i_gps_StreamLexer *gps)
{
  if (mu->next == mu->beyond)
    return mu_fillbuf (mu, gps);

  return mu_buf[mu->next++];
}

tom_int
i_mu_Lexer_i_nextToken (tom_object self, selector cmd)
{
#define BACKUP  do if (c != -1) mu->next--; while (0)

  struct _es_i_gps_SkeletonLexer *gps
    = trt_ext_address (self, _ei_i_gps_SkeletonLexer);
  struct _es_i_mu_Lexer *mu = trt_ext_address (self, _ei_i_mu_Lexer);
  struct _es_i_gps_StreamLexer *sps
    = trt_ext_address (self, _ei_i_gps_StreamLexer);
  tom_int c, i, start_line;

  mu_buf = TRT_SEND ((pointer_imp), mu->buf,
		     USEL (tom, _pi__byteStringContents), &i);
  gps->semantics = 0;

retry:
  mu->start = mu->next;
  c = nextch (mu, sps);
  switch (c)
    {
    case -1:
      gps->token = c_gps_ParserDefines_EOF;
      break;

    case '\n':
      gps->current_line++;
    case ' ':
    case '\t':
      goto retry;

    case '/':
      start_line = gps->current_line;
      c = nextch (mu, sps);

      if (c == '/')
	{
	  /* Absorb double-slash comment.  */
	  do
	    {
	      c = nextch (mu, sps);
	      if (c == -1)
		return gps->token = c_gps_ParserDefines_EOF;
	    } while (c != '\n');
	  gps->current_line++;
	  goto retry;

	  /*NOTREACHED*/
	}

      if (c == '*')
	{
	  /* Absorb C-style comment.  */
	  int b = 0;

	  for (;;)
	    {
	      c = nextch (mu, sps);
	      if (c == -1)
		return mu_premature_eof (mu, gps, start_line, "comment", 7);
	      if (c == '\n')
		gps->current_line++;
	      else if (c == '/')
		{
		  if (b == '*')
		    goto retry;
		}
	      b = c;
	    }

	  /*NOTREACHED*/
	}

      if (c == '=')
	gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_div_is));
      else
	{
	  BACKUP;
	  gps->token = mu->intvalue = '/';
	}
      break;

    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
      {
	int hex = 0, floating = 0, floating_double = -1;
	tom_object n, cst, t;
	tom_long v;
	tom_byte o, is_long = 0;
	tom_int i;

	if (c == '0')
	  {
	    c = nextch (mu, sps);
	    if (c == 'x' || c == 'X')
	      {
		c = nextch (mu, sps);
		hex = 1;
	      }
	  }

	if (hex)
	  while (isdigit (c)
		 || (c >= 'a' && c <= 'f') || (c >= 'A' && c <= 'F'))
	    c = nextch (mu, sps);
	else
	  {
	    while (isdigit (c) || c == 'e' || c == 'E'
		   || c == 'd' || c == 'D' || c == '.'
		   || c == '-' || c == '+')
	      {
		if (!isdigit (c))
		  {
		    floating = 1;
		    if (c == 'd' || c == 'D')
		      floating_double = mu->next - 1;
		  }
		c = nextch (mu, sps);
	      }
	  }

	/* What if more garbage follows the `L'?  */
	if (!floating && (c == 'L' || c == 'l'))
	  is_long = 1;
	else
	  mu->next--;

	if (mu->skips)
	  gps->semantics = 0;
	else
	  {
	    if (floating)
	      {
		int len = mu->next - mu->start;
		double d;

		char *s = alloca (len + 1);
		memcpy (s, mu_buf + mu->start, len);
		s[len] = 0;
		if (floating_double >= 0)
		  s[floating_double - mu->start] = 'e';

		d = atof (s);

		n = TRT_SEND (_PI_, CREF (tom_DoubleNumber), SEL (r_alloc));
		n = TRT_SEND (_PI_, n, USEL (tom, r_init_d), (tom_double) d);
		t = TRT_SEND (_PI_, c_mu_mu_basic_types_from_kind, SEL (r_at_i),
			      (floating_double >= 0 ? c_mu_mu_BT_DOUBLE
			       : c_mu_mu_BT_FLOAT));
	      }
	    else
	      {
		v = TRT_SEND (_II_, mu->buf,
			      USEL (tom, _ioi__integerValue__ii__defaultBase__i_allowSign__o_allowCBases__o_baseSeparator__b_decimalBase__o), mu->start,
			      mu->next - mu->start, 10, 0, 1, '_', 1, &o, &i);

		n = TRT_SEND (_PI_, CREF (tom_IntNumber), SEL (r_alloc));
		n = TRT_SEND (_PI_, n, USEL (tom, r_init_i), (tom_int) v);
		t = TRT_SEND (_PI_, c_mu_mu_basic_types_from_kind, SEL (r_at_i),
			      is_long ? c_mu_mu_BT_LONG : c_mu_mu_BT_INT);
	      }

	    cst = TRT_SEND (_PI_, CREF (mu_NumberConstant), SEL (r_alloc));
	    gps->semantics
	      = TRT_SEND (_PI_, cst, SEL (r_initWithValue_r_type_r), n, t);
	  }
	gps->token = TRT_SEND (, self, SEL (i_number));
	break;
      }

    case 'a': case 'b': case 'c': case 'd': case 'e': case 'f': case 'g':
    case 'h': case 'i': case 'j': case 'k': case 'l': case 'm': case 'n':
    case 'o': case 'p': case 'q': case 'r': case 's': case 't': case 'u':
    case 'v': case 'w': case 'x': case 'y': case 'z':
    case 'A': case 'B': case 'C': case 'D': case 'E': case 'F': case 'G':
    case 'H': case 'I': case 'J': case 'K': case 'L': case 'M': case 'N':
    case 'O': case 'P': case 'Q': case 'R': case 'S': case 'T': case 'U':
    case 'V': case 'W': case 'X': case 'Y': case 'Z':
      do
	{
	  c = nextch (mu, sps);
	} while (mu_identch[c + 1]);
      if (c != -1)
	{
	  do
	    {
	      mu->next--;
	    } while (mu_buf[mu->next - 1] == '-');
	}

      i = mu->next - mu->start;
      gps->semantics = (void *) byte_string_with_string (mu_buf + mu->start, i);
      gps->token = TRT_SEND (, self, SEL (i_identifier));
      break;

    case '"':
      start_line = gps->current_line;
      mu->start++;
      do
	{
	  c = nextch (mu, sps);
	  if (c == '\\')
	    {
	      c = nextch (mu, sps);
	      if (c != -1)
		c = nextch (mu, sps);
	    }
	  if (c == -1)
	    return mu_premature_eof (mu, gps, start_line, "string", 6);
	} while (c != '"');

      if (mu->skips)
	gps->semantics = 0;
      else
	{
	  gps->semantics = TRT_SEND (_PI_, CREF (tom_MutableByteString),
				     SEL (r_withCapacity_i),
				     mu->next - mu->start - 1);
	  while (mu->start != mu->next - 1)
	    {
	      c = mu_buf[mu->start++];
	      if (c == '\\')
		switch (mu_buf[mu->start++])
		  {
		  case -1:
		    return mu_premature_eof (mu, gps, start_line, "string", 6);
		  case 'f': c = '\f'; break;
		  case 'n': c = '\n'; break;
		  case 'r': c = '\r'; break;
		  case 't': c = '\t'; break;
		  case '"': c = '"'; break;
		  case '\\': c = '\\'; break;
		  case '\n': continue;
		  default:
		    /* XXX Unknown escape.  */
		    abort ();
		  }
	      TRT_SEND (, gps->semantics, USEL (tom, v_add_b), c);
	    }
	}
      gps->token = TRT_SEND (, self, SEL (i_string));
      break;

    case '\'':
      {
	int d = nextch (mu, sps);

	start_line = gps->current_line;

	if (d == '\\')
	  {
	    d = nextch (mu, sps);
	    switch (d)
	      {
	      case -1:
		return mu_premature_eof (mu, gps, start_line, "string", 6);
	      case 'f': d = '\f'; break;
	      case 'n': d = '\n'; break;
	      case 'r': d = '\r'; break;
	      case 't': d = '\t'; break;
	      case '\'': d = '\''; break;
	      case '\\': d = '\\'; break;
	      default:
		/* XXX Unknown escape.  */
		abort ();
	      }
	  }
	c = d == -1 ? d : c = nextch (mu, sps);
	if (c == -1)
	  return mu_premature_eof (mu, gps, start_line, "char constant", 13);
	else if (c != '\'')
	  /* XXX Bad character.  */
	  abort ();

	gps->token = TRT_SEND (, self, SEL (i_character));
	if (mu->skips)
	  gps->semantics = 0;
	else
	  {
	    gps->semantics = TRT_SEND (_PI_, CREF (mu_CharConstant),
				       SEL (r_alloc));
	    gps->semantics = TRT_SEND (_PI_, gps->semantics,
				       USEL (tom, r_init_i), d);
	  }
      }
    break;

    case '+':
      c = nextch (mu, sps);
      if (c == '+')
	gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_plusplus));
      else if (c == '=')
	gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_add_is));
      else
	{
	  BACKUP;
	  gps->token = mu->intvalue = '+';
	}
      break;

    case '-':
      c = nextch (mu, sps);
      if (c == '-')
	gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_minusminus));
      else if (c == '>')
	{
	  c = nextch (mu, sps);
	  if (c == '=')
	    gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_implies_is));
	  else
	    {
	      BACKUP;
	      gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_implies));
	    }
	}
      else if (c == '=')
	gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_sub_is));
      else
	{
	  BACKUP;
	  gps->token = mu->intvalue = '-';
	}
      break;

    case '|':
      c = nextch (mu, sps);
      if (c == '|')
	{
	  c = nextch (mu, sps);
	  if (c == '=')
	    gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_sc_or_is));
	  else
	    {
	      BACKUP;
	      gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_oror));
	    }
	}
      else if (c == '{')
	gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_block_open));
      else if (c == '=')
	gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_or_is));
      else
	{
	  BACKUP;
	  gps->token = mu->intvalue = '|';
	}
      break;

    case '}':
      c = nextch (mu, sps);
      if (c == '|')
	gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_block_close));
      else
	{
	  BACKUP;
	  gps->token = mu->intvalue = '}';
	}
      break;

    case '&':
      c = nextch (mu, sps);
      if (c == '&')
	{
	  c = nextch (mu, sps);
	  if (c == '=')
	    gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_sc_and_is));
	  else
	    {
	      BACKUP;
	      gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_andand));
	    }
	}
      else if (c == '=')
	gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_and_is));
      else
	{
	  BACKUP;
	  gps->token = mu->intvalue = '&';
	}
      break;

    case '^':
      c = nextch (mu, sps);
      if (c == '=')
	gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_eor_is));
      else
	{
	  BACKUP;
	  gps->token = mu->intvalue = '^';
	}
      break;

    case '*':
      c = nextch (mu, sps);
      if (c == '=')
	gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_mul_is));
      else
	{
	  BACKUP;
	  gps->token = mu->intvalue = '*';
	}
      break;

    case '!':
      c = nextch (mu, sps);
      if (c == '=')
	gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_ne));
      else
	{
	  BACKUP;
	  gps->token = mu->intvalue = '!';
	}
      break;

    case '=':
      c = nextch (mu, sps);
      if (c == '=')
	gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_eq));
      else
	{
	  BACKUP;
	  gps->token = mu->intvalue = '=';
	}
      break;

    case '>':
      c = nextch (mu, sps);
      if (c == '>')
	{
	  c = nextch (mu, sps);
	  if (c == '>')
	    {
	      c = nextch (mu, sps);
	      if (c == '=')
		gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_lsr_is));
	      else
		{
		  BACKUP;
		  gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_lsr));
		}
	    }
	  else if (c == '=')
	    gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_shr_is));
	  else
	    {
	      BACKUP;
	      gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_shr));
	    }
	}
      else if (c == '=')
	gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_ge));
      else
	{
	  BACKUP;
	  gps->token = mu->intvalue = '>';
	}
      break;

    case '<':
      start_line = gps->current_line;
      mu->start = mu->next;
      do
	{
	  c = nextch (mu, sps);
	} while (c > ' ' && (isalnum (c) || c == '_'));

      if (c != '>')
	{
	  /* This is not a doc comment.  */
	  gps->token = mu->intvalue = '<';

	  if (c != -1)
	    {
	      mu->next = mu->start;

	      c = nextch (mu, sps);
	      if (c == '=')
		gps->token = mu->intvalue = TRT_SEND (, self, SEL (i_le));
	      else if (c == '<')
		{
		  c = nextch (mu, sps);
		  if (c == '=')
		    gps->token = mu->intvalue
		      = TRT_SEND (, self, SEL (i_shl_is));
		  else
		    {
		      BACKUP;
		      gps->token = mu->intvalue
			= TRT_SEND (, self, SEL (i_shl));
		    }
		}
	      else
		mu->next--;
	    }
	  break;
	}

      {
	int tag_len = mu->next - mu->start - 1;

	do
	  {
	    while (c != -1 && c != '<')
	      {
		if (c == '\n')
		  gps->current_line++;
		c = nextch (mu, sps);
	      }

	    if (c == '<')
	      {
		int j;

		c = nextch (mu, sps);
		if (c != '/')
		  {
		    if (c == '\n')
		      gps->current_line++;
		    continue;
		  }

		i = mu->next;
		j = 0;
		do
		  {
		    c = nextch (mu, sps);
		  } while (c == mu_buf[mu->start + j++] && c != '>');

		if (j == tag_len + 1)
		  break;

		continue;
	      }
	  } while (c != -1);

	if (c == -1)
	  return mu_premature_eof (mu, gps, start_line, "documentation", 13);

	if (!mu->skips)
	  {
	    TRT_SEND (_PI_, mu->doc_tag, USEL (tom, r_init__ii__in_r),
		      mu->start, tag_len, mu->buf);
	    if (TRT_SEND (_BI_, self, SEL (o_docp_r), mu->doc_tag))
	      {
		tom_int body_len = mu->next - mu->start - 2 * tag_len - 4;
		tom_int body_start = mu->start + tag_len + 1;
		tom_object tag, body;

		while (isspace (mu_buf[body_start])
		       && mu_buf[body_start] != '\n'
		       && body_len)
		  {
		    body_start++;
		    body_len--;
		  }

		while (body_len && isspace (mu_buf[body_start + body_len - 1])
		       && mu_buf[body_start + body_len - 1] != '\n')
		  body_len--;

		tag = byte_string_with_string (mu_buf + mu->start, tag_len);
		body = byte_string_with_string (mu_buf + body_start, body_len);
		return TRT_SEND (_II_, self, SEL (i_documentation_r_tag_r_at_i),
				 body, tag, start_line);
	      }
	  }
      }
      goto retry;

    default:
      gps->token = mu->intvalue = c;
      break;
    }

  return gps->token;
#undef BACKUP
}
